MODULE cxios

USE PARKIND1, ONLY : JPIM, JPRM, JPRB
USE YOMXIOS
USE XIOS

IMPLICIT NONE
PRIVATE

PUBLIC :: IFS_XIOS_CALENDAR, XIOS_PUT, XIOS_FLUSH_BUFFERS

CONTAINS

SUBROUTINE IFS_XIOS_CALENDAR

  USE YOMCT3, ONLY : NSTEP
  ! XIOS_FPOS extra logging
  USE YOMLUN , ONLY : NULOUT

  !$OMP SINGLE
  ! Update calendar of XIOS with the current time step
  IF ((NSTEP-nstep_from_origin) > 0) THEN
    CALL XIOS_UPDATE_CALENDAR(NSTEP-nstep_from_origin)
    WRITE(NULOUT, '(''XIOSFPOS UPDATE CALENDAR NSTEP-CSTEP IS'',I8)') NSTEP-nstep_from_origin
    WRITE(NULOUT, '(''XIOSFPOS UPDATE CALENDAR NSTEP IS'',I8)') NSTEP
  END IF
  !$OMP END SINGLE

END SUBROUTINE IFS_XIOS_CALENDAR

SUBROUTINE XIOS_PUT(YDGEOMETRY,YDR,CMODE,PR2,PR3,PR4,KR2,KR3,KR4)

!**** *XIOS_PUT*  - Send data to XIOS

! Purpose - To send fields to output server

! Explicit arguments :
! --------------------
! YDGEOMETRY - Global geometry structure
! YDR   - IO request structure
! CMODE - read/write/append mode. Used to differentiate non-lagged ('w') and lagged ('a') surface fields
! PR2   - real rank 2 array
! PR3   - real rank 3 array
! PR4   - real rank 4 array
! KR2   - integer rank 2 array (currently ignored)
! KR3   - integer rank 3 array (currently ignored)
! KR4   - integer rank 4 array (currently ignored)
!-------------------------------------------------------------------------

  USE GEOMETRY_MOD, ONLY: GEOMETRY
  USE IOSTREAM_MIX, ONLY: TYPE_IOREQUEST,GET_PUT_ARGS,JPGP,JPSPEC

  IMPLICIT NONE
  
  TYPE(GEOMETRY)       , INTENT(IN)           :: YDGEOMETRY
  TYPE(TYPE_IOREQUEST) , INTENT(IN)           :: YDR
  CHARACTER(LEN=1)     , INTENT(IN)           :: CMODE
  REAL(KIND=JPRB)      , OPTIONAL, INTENT(IN) :: PR2(:,:)       ! 1 2D gp field
                                                                ! 1 3D sp field
                                                                ! n 2D sp fields
  REAL(KIND=JPRB)      , OPTIONAL, INTENT(IN) :: PR3(:,:,:)     ! 1 3D gp field
                                                                ! n 2D gp fields
                                                                ! n 3D sp fields
  REAL(KIND=JPRB)      , OPTIONAL, INTENT(IN) :: PR4(:,:,:,:)   ! n 3D gp fields
  INTEGER(KIND=JPIM)   , OPTIONAL, INTENT(IN) :: KR2(:,:)
  INTEGER(KIND=JPIM)   , OPTIONAL, INTENT(IN) :: KR3(:,:,:)
  INTEGER(KIND=JPIM)   , OPTIONAL, INTENT(IN) :: KR4(:,:,:,:)
  
  LOGICAL :: LLPR(4),LLKR(4)
  INTEGER(KIND=JPIM) :: IPRD(4,4),IKRD(4,4)

  CALL GET_PUT_ARGS(YDR,LLPR,IPRD,LLKR,IKRD,PR2=PR2,PR3=PR3,PR4=PR4,KR2=KR2,KR3=KR3,KR4=KR4)
  SELECT CASE(YDR%IDATA_TYPE)
    CASE(JPGP)
      IF(LLPR(1)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 1 GRIDPOINT DATA IN XIOS OUTPUT STREAM'
      ENDIF
      IF(LLPR(2)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 2 GRIDPOINT DATA IN XIOS OUTPUT STREAM'
      ENDIF
      IF(LLPR(3)) THEN
        CALL XIOS_FLD_OUT(YDGEOMETRY,YDR,CMODE,PR3=PR3,NFLDS=IPRD(3,2))
      ENDIF
      IF(LLPR(4)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 4 GRIDPOINT DATA IN XIOS OUTPUT STREAM'
      ENDIF
    CASE(JPSPEC)
      IF(LLPR(1)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 1 SPECTRAL DATA IN XIOS OUTPUT STREAM'
      ENDIF
      IF(LLPR(2)) THEN
        CALL XIOS_FLD_OUT(YDGEOMETRY,YDR,CMODE,PR2=PR2,NFLDS=YDR%YMDL_2D%NUMFLDS)
      ENDIF
      IF(LLPR(3)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 3 SPECTRAL DATA IN XIOS OUTPUT STREAM'
      ENDIF
      IF(LLPR(4)) THEN
        WRITE(0,*) 'UNSUPPORTED RANK 4 SPECTRAL DATA IN XIOS OUTPUT STREAM'
      ENDIF
    CASE DEFAULT
      WRITE(0,*) 'UNSUPPORTED DATA TYPE IN XIOS STREAM:',YDR%IDATA_TYPE
  END SELECT

END SUBROUTINE XIOS_PUT

SUBROUTINE XIOS_FLD_OUT(YDGEOMETRY,YDR,CMODE,PR2,PR3,NFLDS)

!**** *XIOS_FLD_OUT*  - Output data to XIOS

! Purpose - Gather and send fields to XIOS

! Explicit arguments :
! --------------------
! YDGEOMETRY - Geometry data structure
! YDR     - IO request structure
! CMODE - read/write/append mode. Used to differentiate non-lagged ('w') and lagged ('a') surface fields
! PR2     - 2-D array: array of sp fields
! PR3     - 3-D array: array of gp fields
!-------------------------------------------------------------------------
  USE GEOMETRY_MOD ,    ONLY : GEOMETRY
  USE IOSTREAM_MIX ,    ONLY : TYPE_IOREQUEST,YIOS
  USE YOMXIOS,          ONLY : NSFCFLD,CSFCFLD,IGRBSFCFLD
  USE YOMFPC,           ONLY : NFP3S,NFP3P,NFP3PV,NFP3TH
  USE YOMFPDIM ,        ONLY : NFPISCAG
  ! XIOS_FPOS extra logging
  USE YOMLUN          , ONLY : NULOUT

  IMPLICIT NONE
    
  TYPE(GEOMETRY)      , INTENT(IN)          :: YDGEOMETRY
  TYPE(TYPE_IOREQUEST), INTENT(IN)          :: YDR
  CHARACTER(LEN=1)    , INTENT(IN)          :: CMODE
  REAL(KIND=JPRB)     , OPTIONAL,INTENT(IN) :: PR2(:,:)
  REAL(KIND=JPRB)     , OPTIONAL,INTENT(IN) :: PR3(:,:,:)
  INTEGER             , INTENT(IN)          :: NFLDS

  REAL(KIND=JPRB), ALLOCATABLE :: PXIOSBF2(:,:)
  REAL(KIND=JPRB), ALLOCATABLE :: PXIOSBF3(:,:,:)
  CHARACTER(LEN=3) :: CPOST
  CHARACTER(LEN=16) :: CGRB
  INTEGER(KIND=JPIM) :: IGRB,JFLD,STEP,JSTGLO,JLEV,ICEND,IBL,I,J,KLEVS,KGPTOT,KPROMA
  LOGICAL            :: CODEFOUND

#include "trans_inq.h"
#include "inv_trans.h"
  
  KLEVS=1
  CPOST='   '
  WRITE(NULOUT, '(''XIOSFPOS: LEVEL TYPE IS'',A4)') YDR%CLEVTYPE
  IF (YDR%CLEVTYPE.EQ.'SFC') THEN
      KLEVS=1
  ELSEIF (YDR%CLEVTYPE.EQ.'ML') THEN
      KLEVS=NFP3S
  ELSEIF (YDR%CLEVTYPE.EQ.'PL') THEN
      KLEVS=MAX(NFP3P,1) ! Necessary for sp & lnsp only data blocks
      CPOST='_pl'
  ELSEIF (YDR%CLEVTYPE.EQ.'PV') THEN
      KLEVS=NFP3PV
      CPOST='_pv'
  ELSEIF (YDR%CLEVTYPE.EQ.'TH') THEN
      KLEVS=NFP3TH
      CPOST='_th'
  ENDIF
  
  CALL TRANS_INQ(KRESOL=YDR%NRESOL,KGPTOT=KGPTOT)

  KPROMA=YDR%YGPF%NPROMA

  WRITE(NULOUT, '(''XIOSFPOS: KGPTOT IS'',I8)') KGPTOT
  WRITE(NULOUT, '(''XIOSFPOS: KLEVS IS'',I8)') KLEVS
  WRITE(NULOUT, '(''XIOSFPOS: NFLDS IS'',I8)') NFLDS
  WRITE(NULOUT, '(''XIOSFPOS: KPROMA IS'',I8)') KPROMA
  IF(PRESENT(PR2)) THEN
    WRITE(NULOUT, '(''XIOSFPOS: FIRST DIMENSION OF PR2 IS'',I8)') UBOUND(PR2,1)
    WRITE(NULOUT, '(''XIOSFPOS: SECOND DIMENSION OF PR2 IS'',I8)') UBOUND(PR2,2)
    WRITE(NULOUT, '(''XIOSFPOS: NFPISCAG IS'',I8)') NFPISCAG
  ENDIF
  CALL FLUSH(NULOUT)

  IF(PRESENT(PR2)) THEN
    ALLOCATE(PXIOSBF3(KGPTOT,NFPISCAG,1))
    WRITE(NULOUT, '(''XIOSFPOS: INVERSE SPECTRAL TRANSFORM'')')
    CALL INV_TRANS(PSPSCALAR=TRANSPOSE(PR2),KPROMA=KGPTOT,KVSETSC=YDR%YMDL_2D%IBSET(:),&
     & KRESOL=YDR%NRESOL,PGP=PXIOSBF3)
  ELSE
    ALLOCATE(PXIOSBF2(KGPTOT,KLEVS))
  ENDIF

  !$OMP SINGLE

  JFLD=1
  STEP=0
  DO WHILE(JFLD<=NFLDS)
    IGRB=YDR%YMDL_2D%IGRIBCODE(JFLD)
    WRITE(NULOUT, '(''XIOSFPOS: GRIB CODE IS'',I8)') IGRB
    CODEFOUND=.FALSE.
    IF (YDR%CLEVTYPE.EQ.'SFC') THEN
      DO I=1,NSFCFLD
        IF (IGRBSFCFLD(I).EQ.IGRB) THEN
          CGRB=CSFCFLD(I)
          CODEFOUND=.TRUE.
          EXIT
        ENDIF
      ENDDO
    ELSE
      DO I=1,N3DFLD
        IF (IGRB3DFLD(I).EQ.IGRB) THEN
          CGRB=TRIM(C3DFLD(I))//CPOST
          CODEFOUND=.TRUE.
          EXIT
        ENDIF
      ENDDO
      IF (.NOT.CODEFOUND) THEN ! Search for PEXTRA fields
        DO I=1,NPEXTRAFLD
          IF (IGRBPEXTRAFLD(I).EQ.IGRB) THEN
            CGRB=TRIM(CPEXTRAFLD(I))
            CODEFOUND=.TRUE.
            EXIT
          END IF
        END DO
      END IF
    ENDIF
    IF (.NOT.CODEFOUND) THEN
        ! 3D spectral blocks always end with a slice of surface pressure-> skip
        ! it
        IF (YDR%CLEVTYPE.NE.'SFC'.AND.IGRB.EQ.134) THEN
          STEP=1
          JFLD=JFLD+STEP
          CYCLE
        ELSE IF (YDR%CLEVTYPE.EQ.'PL'.AND.IGRB.EQ.152) THEN
          STEP=1
          CGRB='lnsp            '
        ELSE
          STEP=KLEVS
          JFLD=JFLD+STEP
          IF (IGRB.NE.134.AND.IGRB.NE.152) WRITE(NULOUT, '(''XIOSFPOS: UNKNOWN GRIB CODE SKIPPED'',I8)') IGRB
          CYCLE
        ENDIF
    ELSE
        ! Skipped surface geopotential field at the end of model levels fields
        IF (IGRB.EQ.129.AND.(JFLD+KLEVS-1>NFLDS)) THEN
          STEP=1
          JFLD=JFLD+STEP
          CYCLE
        ENDIF
        STEP=KLEVS
    ENDIF
    IF (LOPT_SEND) THEN
      IF(PRESENT(PR2)) THEN
        IF (LSINGLE_PREC_SEND) THEN
          WRITE(NULOUT, '(''XIOSFPOS: COPYING SP BLOCK TO DELAYED BUFFER IN SINGLE PRECISION'')')
          CALL XIOS_FILL_FIELD_BUFFER(FLDTYPE=YDR%CLEVTYPE, CGRB=CGRB, FLDBUF_SP=REAL(PXIOSBF3(:,JFLD:JFLD+STEP-1,1),JPRM))
        ELSE
          WRITE(NULOUT, '(''XIOSFPOS: COPYING SP BLOCK TO DELAYED BUFFER IN DOUBLE PRECISION'')')
          CALL XIOS_FILL_FIELD_BUFFER(FLDTYPE=YDR%CLEVTYPE, CGRB=CGRB, FLDBUF_DP=PXIOSBF3(:,JFLD:JFLD+STEP-1,1))
        END IF
      ELSEIF(PRESENT(PR3)) THEN
        DO JSTGLO=1,KGPTOT,KPROMA
          ICEND=MIN(KPROMA,KGPTOT-JSTGLO+1)
          IBL=(JSTGLO-1)/KPROMA+1
          DO JLEV=1,STEP
            I=JSTGLO
            DO J=1,ICEND
              PXIOSBF2(I,JLEV)=PR3(J,JFLD+JLEV-1,IBL)
              I=I+1
            ENDDO
          ENDDO
        ENDDO
        IF (YDR%CLEVTYPE == 'SFC' .AND. CMODE == 'a') THEN ! Lagged surface fields are sent immediately
          WRITE(NULOUT, '(''XIOSFPOS: CHECKING FOR ACTIVE '',A16)') TRIM(CGRB)
          IF(XIOS_FIELD_IS_ACTIVE(TRIM(CGRB),.TRUE.)) THEN
            IF (LSINGLE_PREC_SEND) THEN
              WRITE(NULOUT, '(''XIOSFPOS: SENDING GP BLOCK TO XIOS IN SINGLE PRECISION'')')
              CALL XIOS_SEND_FIELD(TRIM(CGRB),REAL(PXIOSBF2(:,:),JPRM))
            ELSE
              WRITE(NULOUT, '(''XIOSFPOS: SENDING GP BLOCK TO XIOS IN DOUBLE PRECISION'')')
              CALL XIOS_SEND_FIELD(TRIM(CGRB),PXIOSBF2(:,:))
            END IF
          END IF
        ELSE
          IF (LSINGLE_PREC_SEND) THEN
            WRITE(NULOUT, '(''XIOSFPOS: COPYING GP BLOCK TO DELAYED BUFFER IN SINGLE PRECISION'')')
            CALL XIOS_FILL_FIELD_BUFFER(FLDTYPE=YDR%CLEVTYPE, CGRB=CGRB, FLDBUF_SP=REAL(PXIOSBF2(:,:),JPRM))
          ELSE
            WRITE(NULOUT, '(''XIOSFPOS: COPYING GP BLOCK TO DELAYED BUFFER IN DOUBLE PRECISION'')')
            CALL XIOS_FILL_FIELD_BUFFER(FLDTYPE=YDR%CLEVTYPE, CGRB=CGRB, FLDBUF_DP=PXIOSBF2(:,:))
          END IF
        END IF
      ENDIF
    ELSE
      WRITE(NULOUT, '(''XIOSFPOS: CHECKING FOR ACTIVE '',A16)') TRIM(CGRB)
      IF(XIOS_FIELD_IS_ACTIVE(TRIM(CGRB),.TRUE.)) THEN
        IF(PRESENT(PR2)) THEN
          IF (LSINGLE_PREC_SEND) THEN
            WRITE(NULOUT, '(''XIOSFPOS: SENDING SP BLOCK TO XIOS IN SINGLE PRECISION'')')
            CALL XIOS_SEND_FIELD(TRIM(CGRB),REAL(PXIOSBF3(:,JFLD:JFLD+STEP-1,1),JPRM))
          ELSE
            WRITE(NULOUT, '(''XIOSFPOS: SENDING SP BLOCK TO XIOS IN DOUBLE PRECISION'')')
            CALL XIOS_SEND_FIELD(TRIM(CGRB),PXIOSBF3(:,JFLD:JFLD+STEP-1,1))
          END IF
        ELSEIF(PRESENT(PR3)) THEN
          DO JSTGLO=1,KGPTOT,KPROMA
            ICEND=MIN(KPROMA,KGPTOT-JSTGLO+1)
            IBL=(JSTGLO-1)/KPROMA+1
            DO JLEV=1,STEP
              I=JSTGLO
              DO J=1,ICEND
                PXIOSBF2(I,JLEV)=PR3(J,JFLD+JLEV-1,IBL)
                I=I+1
              ENDDO
            ENDDO
          ENDDO
          IF (LSINGLE_PREC_SEND) THEN
            WRITE(NULOUT, '(''XIOSFPOS: SENDING GP BLOCK TO XIOS IN SINGLE PRECISION'')')
            CALL XIOS_SEND_FIELD(TRIM(CGRB),REAL(PXIOSBF2(:,:),JPRM))
          ELSE
            WRITE(NULOUT, '(''XIOSFPOS: SENDING GP BLOCK TO XIOS IN DOUBLE PRECISION'')')
            CALL XIOS_SEND_FIELD(TRIM(CGRB),PXIOSBF2(:,:))
          END IF
        ENDIF
      ENDIF
    END IF
    JFLD=JFLD+STEP
  ENDDO

  !$OMP END SINGLE

  IF (ALLOCATED(PXIOSBF2)) DEALLOCATE(PXIOSBF2)
  IF (ALLOCATED(PXIOSBF3)) DEALLOCATE(PXIOSBF3)

END SUBROUTINE XIOS_FLD_OUT

SUBROUTINE XIOS_FILL_FIELD_BUFFER(FLDTYPE, CGRB, FLDBUF_SP, FLDBUF_DP)

  ! XIOS_FPOS extra logging
  USE YOMLUN, ONLY : NULOUT

  CHARACTER(LEN=*),  INTENT(IN) :: FLDTYPE
  CHARACTER(LEN=16), INTENT(IN) :: CGRB
  REAL(KIND=JPRM), OPTIONAL, INTENT(IN) :: FLDBUF_SP(:,:)
  REAL(KIND=JPRB), OPTIONAL, INTENT(IN) :: FLDBUF_DP(:,:)
  
  IF (TRIM(FLDTYPE) == 'PL' .AND. TRIM(CGRB) == TRIM('lnsp')) THEN ! Logarithm of surface pressure (lnsp)
    NSFCFLDBUF=NSFCFLDBUF+1
    CSFCFLDBUF(NSFCFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      SFCFLDBUF_SP(:,NSFCFLDBUF)=FLDBUF_SP(:,1)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      SFCFLDBUF_DP(:,NSFCFLDBUF)=FLDBUF_DP(:,1)
    END IF
  ELSE IF (TRIM(FLDTYPE) == 'SFC') THEN ! Surface physical fields
    NSFCFLDBUF=NSFCFLDBUF+1
    CSFCFLDBUF(NSFCFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      SFCFLDBUF_SP(:,NSFCFLDBUF)=FLDBUF_SP(:,1)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      SFCFLDBUF_DP(:,NSFCFLDBUF)=FLDBUF_DP(:,1)
    END IF
  ELSE IF (TRIM(FLDTYPE) == 'ML') THEN ! Model levels
    NMLFLDBUF=NMLFLDBUF+1
    CMLFLDBUF(NMLFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      MLFLDBUF_SP(:,:,NMLFLDBUF)=FLDBUF_SP(:,:)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      MLFLDBUF_DP(:,:,NMLFLDBUF)=FLDBUF_DP(:,:)
    END IF
  ELSE IF (TRIM(FLDTYPE) == 'PL') THEN ! Pressure levels
    NPLFLDBUF=NPLFLDBUF+1
    CPLFLDBUF(NPLFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      PLFLDBUF_SP(:,:,NPLFLDBUF)=FLDBUF_SP(:,:)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      PLFLDBUF_DP(:,:,NPLFLDBUF)=FLDBUF_DP(:,:)
    END IF
  ELSE IF (TRIM(FLDTYPE) == 'TH') THEN ! Theta levels
    NTLFLDBUF=NTLFLDBUF+1
    CTLFLDBUF(NTLFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      TLFLDBUF_SP(:,:,NTLFLDBUF)=FLDBUF_SP(:,:)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      TLFLDBUF_DP(:,:,NTLFLDBUF)=FLDBUF_DP(:,:)
    END IF
  ELSE IF (TRIM(FLDTYPE) == 'PV') THEN ! PV levels
    NVLFLDBUF=NVLFLDBUF+1
    CVLFLDBUF(NVLFLDBUF)=CGRB
    IF(PRESENT(FLDBUF_SP)) THEN
      VLFLDBUF_SP(:,:,NVLFLDBUF)=FLDBUF_SP(:,:)
    ELSE IF (PRESENT(FLDBUF_DP)) THEN
      VLFLDBUF_DP(:,:,NVLFLDBUF)=FLDBUF_DP(:,:)
    END IF
  END IF

END SUBROUTINE XIOS_FILL_FIELD_BUFFER

SUBROUTINE XIOS_FLUSH_BUFFERS

  ! XIOS_FPOS extra logging
  USE YOMLUN , ONLY : NULOUT

  !$OMP SINGLE

  ! Send surface fields
  DO WHILE (NSFCFLDBUF > 0)
    IF(XIOS_FIELD_IS_ACTIVE(TRIM(CSFCFLDBUF(NSFCFLDBUF)),.TRUE.)) THEN
      WRITE(NULOUT, '(''XIOSFPOS: SENDING DELAYED SURFACE FIELD '',A16)') TRIM(CSFCFLDBUF(NSFCFLDBUF))
      IF (LSINGLE_PREC_SEND) THEN
        CALL XIOS_SEND_FIELD(TRIM(CSFCFLDBUF(NSFCFLDBUF)),SFCFLDBUF_SP(:,NSFCFLDBUF))
      ELSE
        CALL XIOS_SEND_FIELD(TRIM(CSFCFLDBUF(NSFCFLDBUF)),SFCFLDBUF_DP(:,NSFCFLDBUF))
      END IF
    END IF
    NSFCFLDBUF=NSFCFLDBUF-1
  END DO

  ! Send model levels fields
  DO WHILE (NMLFLDBUF > 0)
    IF(XIOS_FIELD_IS_ACTIVE(TRIM(CMLFLDBUF(NMLFLDBUF)),.TRUE.)) THEN
      WRITE(NULOUT, '(''XIOSFPOS: SENDING DELAYED MODEL LEVELS FIELD '',A16)') TRIM(CMLFLDBUF(NMLFLDBUF))
      IF (LSINGLE_PREC_SEND) THEN
        CALL XIOS_SEND_FIELD(TRIM(CMLFLDBUF(NMLFLDBUF)),MLFLDBUF_SP(:,:,NMLFLDBUF))
      ELSE
        CALL XIOS_SEND_FIELD(TRIM(CMLFLDBUF(NMLFLDBUF)),MLFLDBUF_DP(:,:,NMLFLDBUF))
      END IF
    END IF
    NMLFLDBUF=NMLFLDBUF-1
  END DO

  ! Send pressure levels fields
  DO WHILE (NPLFLDBUF > 0)
    IF(XIOS_FIELD_IS_ACTIVE(TRIM(CPLFLDBUF(NPLFLDBUF)),.TRUE.)) THEN
      WRITE(NULOUT, '(''XIOSFPOS: SENDING DELAYED PRESSURE LEVELS FIELD '',A16)') TRIM(CPLFLDBUF(NPLFLDBUF))
      IF (LSINGLE_PREC_SEND) THEN
        CALL XIOS_SEND_FIELD(TRIM(CPLFLDBUF(NPLFLDBUF)),PLFLDBUF_SP(:,:,NPLFLDBUF))
      ELSE
        CALL XIOS_SEND_FIELD(TRIM(CPLFLDBUF(NPLFLDBUF)),PLFLDBUF_DP(:,:,NPLFLDBUF))
      END IF
    END IF
    NPLFLDBUF=NPLFLDBUF-1
  END DO

  ! Send theta levels fields
  DO WHILE (NTLFLDBUF > 0)
    IF(XIOS_FIELD_IS_ACTIVE(TRIM(CTLFLDBUF(NTLFLDBUF)),.TRUE.)) THEN
      WRITE(NULOUT, '(''XIOSFPOS: SENDING DELAYED THETA LEVELS FIELD '',A16)') TRIM(CTLFLDBUF(NTLFLDBUF))
      IF (LSINGLE_PREC_SEND) THEN
        CALL XIOS_SEND_FIELD(TRIM(CTLFLDBUF(NTLFLDBUF)),TLFLDBUF_SP(:,:,NTLFLDBUF))
      ELSE
        CALL XIOS_SEND_FIELD(TRIM(CTLFLDBUF(NTLFLDBUF)),TLFLDBUF_DP(:,:,NTLFLDBUF))
      END IF
    END IF
    NTLFLDBUF=NTLFLDBUF-1
  END DO

  ! Send PV levels fields
  DO WHILE (NVLFLDBUF > 0)
    IF(XIOS_FIELD_IS_ACTIVE(TRIM(CVLFLDBUF(NVLFLDBUF)),.TRUE.)) THEN
      WRITE(NULOUT, '(''XIOSFPOS: SENDING DELAYED PV LEVELS FIELD '',A16)') TRIM(CVLFLDBUF(NVLFLDBUF))
      IF (LSINGLE_PREC_SEND) THEN
        CALL XIOS_SEND_FIELD(TRIM(CVLFLDBUF(NVLFLDBUF)),VLFLDBUF_SP(:,:,NVLFLDBUF))
      ELSE
        CALL XIOS_SEND_FIELD(TRIM(CVLFLDBUF(NVLFLDBUF)),VLFLDBUF_DP(:,:,NVLFLDBUF))
      END IF
    END IF
    NVLFLDBUF=NVLFLDBUF-1
  END DO

  !$OMP END SINGLE

END SUBROUTINE XIOS_FLUSH_BUFFERS

END MODULE cxios
